/*
 * This file is part of xsoundtrack - a sound tracker for Linux.
 * Sed - july 1998.
 * This is total free software.
 *-------------------------------------
 * The file contains soundcard-control routine (OSS)
 */

#include "card.h"

#ifdef __linux__
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>

int fragment_size[11]={         /* 7, ..., 17 makes 11 values */
  /* fragment size is in [7..17] says mikmod 3.0.3, so... */
  128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768, 65536, 131072
};

/* the "/dev/dsp" file handle */
int soundd;

static int inited=0;

int get_sound_fd(void)
{
  if (!inited)
    return -1;

  return soundd;
}

/* return 0 if full (or error) 1 if not */
int not_full_sound(void)
{
  audio_buf_info info;

  if (ioctl(soundd, SNDCTL_DSP_GETOSPACE, &info)==-1)
    return 0;

  return info.fragments!=0;
}

/* this function sets the card output resolution, ie 8/16 bits, mono(0)/stereo(1), sampling rate (8000, 44100, ...) */
/* return 0 on error, 1 on success */
int set_sound_type(int bit_resolution, int mono_stereo, int sampling_rate)
{
  int fragsize=12;
  int numfrags=1;

  int fragmentsize=(numfrags<<16) | fragsize;
  if(ioctl(soundd, SNDCTL_DSP_SETFRAGMENT, &fragmentsize)<0)
    return 0;

  if (ioctl(soundd, SNDCTL_DSP_SAMPLESIZE, &bit_resolution)==-1 ||
      ioctl(soundd, SNDCTL_DSP_STEREO, &mono_stereo)==-1 ||
      ioctl(soundd, SNDCTL_DSP_SPEED, &sampling_rate)==-1)
    return 0;
  return 1;
}

/* return 0 if ok, -1 otherwise */
int unblock_sound(int fd)
{
  int fl;

  if ((fl=fcntl(fd, F_GETFL))==-1) {
    perror("card.unblock_sound.fcntl1");
    return -1;
  }
  fl|=O_NONBLOCK;
  if ((fl=fcntl(fd, F_SETFL, fl))==-1) {
    perror("card.unblock_soud.fcntl2");
    return -1;
  }
  return 0;
}

/* return 0 on error 1 otherwise*/
int init_sound(int sample_rate, int precision, int mono_stereo)
{
  if (inited==1)
    return 1;
  soundd=open("/dev/dsp", O_WRONLY, 0);
  if (soundd==-1 || !set_sound_type(precision, mono_stereo, sample_rate) /*|| unblock_sound(soundd)==-1*/) {
    if (soundd!=-1)
      close(soundd);
    return 0;
  }
  inited=1;

  /* we put some mixer values (should be an option, will see that */
  /* don't care if a problem occurs */
  if (0) {
    int outvol=90*256+90;
    int mixfd=open("/dev/mixer", O_WRONLY);
    if (!mixfd)
      return 1;
    /* set volume output to 90 */
    if (ioctl(mixfd, MIXER_WRITE(0), &outvol)==-1) {
      close(mixfd);
      return 1;
    }
    /* set bass to 99 */
    outvol=99*256+99;
    if (ioctl(mixfd, MIXER_WRITE(1), &outvol)==-1) {
      close(mixfd);
      return 1;
    }
    /* set trebl to 82 */
    outvol=82*256+82;
    if (ioctl(mixfd, MIXER_WRITE(2), &outvol)==-1) {
      close(mixfd);
      return 1;
    }
    /* set output gain to 75 */
    outvol=75*256+75;
    if (ioctl(mixfd, MIXER_WRITE(13), &outvol)==-1) {
      close(mixfd);
      return 1;
    }
    /* set input gain to 75 */
    outvol=75*256+75;
    if (ioctl(mixfd, MIXER_WRITE(12), &outvol)==-1) {
      close(mixfd);
      return 1;
    }
    /* now the microphone volume to 0 and line volume too */
    outvol=0;
    if (ioctl(mixfd, MIXER_WRITE(7), &outvol)==-1) {
      close(mixfd);
      return 1;
    }
    if (ioctl(mixfd, MIXER_WRITE(6), &outvol)==-1) {
      close(mixfd);
      return 1;
    }
    close(mixfd);
  }

  return 1;
}

void close_sound(void)
{
  if (!inited)
    return;
  close(soundd);
  inited=0;
}

/* return 0 if problem, 1 if ok */
int play_sound(void *buf, int buf_len)
{
  if (!inited)
    return 0;
  if (write(soundd, buf, buf_len)==-1)
    return 0;
  return 1;
}

/* recording functions */
#define recfd soundd
int buf_size;
int buf_size_inc;

/* return 0 on error 1 if ok */
int open_record(int rate, int _16_bits, int frag_size, int nb_frags, int line_micro, int input_volume)
{
  int fragmentsize=(nb_frags<<16) | frag_size;
  int mono_stereo=0;
  extern int fragment_size[11];

  if (_16_bits)
    _16_bits=16;
  else
    _16_bits=8;

  recfd=open("/dev/dsp", O_RDONLY, 0);
  if (recfd==-1) {
    perror("card.open_record open");
    return 0;
  }

  if(ioctl(recfd, SNDCTL_DSP_SETFRAGMENT, &fragmentsize)<0)
    goto error;

  if (ioctl(recfd, SNDCTL_DSP_SAMPLESIZE, &_16_bits)==-1 ||
      ioctl(recfd, SNDCTL_DSP_STEREO, &mono_stereo)==-1 ||
      ioctl(recfd, SNDCTL_DSP_SPEED, &rate)==-1)
    goto error;

  if (unblock_sound(recfd)==-1)
    goto error;

  /* now, we gonna close the output volume, and put the right input volume */
  /* the device number are taken from SoundStudio. Thanks to the author of this good tool (even if a lot buggy) */
  {
    int outvol=0, z=0;
    int mixfd=open("/dev/mixer", O_WRONLY);
    if (!mixfd)
      goto error;
    /* set volume output to 0 */
    if (ioctl(mixfd, MIXER_WRITE(0), &outvol)==-1) {
      close(mixfd);
      goto error;
    }
    /* now the microphone volume */
    outvol=input_volume;
    outvol<<=8;
    outvol|=input_volume;
    if (ioctl(mixfd, MIXER_WRITE(7), line_micro?&z:&outvol)==-1) {
      close(mixfd);
      goto error;
    }
    /* now the line volume */
    if (ioctl(mixfd, MIXER_WRITE(6), line_micro?&outvol:&z)==-1) {
      close(mixfd);
      goto error;
    }
    close(mixfd);
  }

  buf_size=0;
  buf_size_inc=fragment_size[frag_size-7];

  return 1;

error:
  perror("card.open_record error");
  close(recfd);
  return 0;
}

/* return the new buf if ok, 0 otherwise (this fun realloc) */
char *read_record(char *buf)
{
  char *dum;

  if (!(dum=realloc(buf, buf_size+buf_size_inc)))
    return (char *)0;

  if (read(recfd, &dum[buf_size], buf_size_inc)!=buf_size_inc)
    perror("card.read_record error in reading");

  buf_size+=buf_size_inc;

  return dum;
}

int get_record_buf_size(void)
{
  return buf_size;
}

void close_record(void)
{
  close(recfd);

  {
    int outvol=90*256+90;
    int mixfd=open("/dev/mixer", O_RDONLY);
    if (!mixfd)
      return;
    /* set volume output to 90 */
    if (ioctl(mixfd, MIXER_WRITE(0), &outvol)==-1) {
      close(mixfd);
      return;
    }
    /* now the microphone volume to 0 and line volume too */
    outvol=0;
    if (ioctl(mixfd, MIXER_WRITE(7), &outvol)==-1) {
      close(mixfd);
      return;
    }
    if (ioctl(mixfd, MIXER_WRITE(6), &outvol)==-1) {
      close(mixfd);
      return;
    }
    close(mixfd);
  }
}
#else
int init_sound(int a, int b, int c)
{
  return 0;
}

void close_sound(void)
{
}

int play_sound(void *a, int b)
{
  return 0;
}

int get_sound_fd(void)
{
  return 0;
}

int open_record(int a, int b, int c, int d, int e, int f)
{
  return 0;
}

void close_record(void)
{
}

char *read_record(char *a)
{
  return 0;
}

int get_record_buf_size(void)
{
  return 0;
}

#endif
