/*
 * the oa project.
 *
 * Sed - december 1999
 *
 * This is public domain.
 *
 * the main file
 */

/* a wave form is defined by control points
 * x goes from 0 to what you want
 * y goes from -32768 to 32767
 * x_size is the width of the wave form (the wave form will repeat at x+x_size)
 * we must have at least 2 control points or all the stuff will crash
 *
 * When you will ask for the sample number n, the x stuff will determine
 * in which interval remains the sample and the y stuff will give you it's value
 * by a simple interpolation (linear or sin or what you want (sin is better, closer
 * to reality))
 * The last interval is between the last control point and the first one.
 */

#include <stdio.h>
#include <X11/Xlib.h>

#include "waves.h"
#include "screen.h"
#include "event.h"

#if 0
int o_x[]={0, 32, 54, 64+32};
int o_y[]={32600, -17000, 17000, -32600};
wave_form wave_o = {
  control_points_x:o_x,
  control_points_y:o_y,
  nb_control_points:4,
  x_size:128
};

int a_x[]={0, 16, 32+10, 64, 70+2, 76, 82-2, 88};
int a_y[]={25000, -15000, 30000, -25000, -15000, -18500, -16000, -30000};
wave_form wave_a = {
  control_points_x:a_x,
  control_points_y:a_y,
  nb_control_points:8,
  x_size:128
};

#define N (8192)
int percent_x[]={0, N/2};
int percent_y[]={0, 100};
wave_form percent = {
  control_points_x:percent_x,
  control_points_y:percent_y,
  nb_control_points:2,
  x_size:N
};

#define M (16384*2)
int intens_x[]={0, M/200};
int intens_y[]={100, 100};
wave_form intens = {
  control_points_x:intens_x,
  control_points_y:intens_y,
  nb_control_points:2,
  x_size:M
};

int get_sample_linear(wave_form *w, int sample_nb)
{
  int y;
  int p1, p2, vx=0;
  float x1, x2, y1, y2;

  /* take the sample in the wave form */
  sample_nb %= w->x_size;

  /* search the good interval */
  p2=0;
  while (p2!=w->nb_control_points && w->control_points_x[p2]<=sample_nb) /* order of test is important */
    p2++;
  p1=p2-1;
  if (p2==w->nb_control_points) {
    p1=w->nb_control_points-1;
    p2=0;
    vx=w->x_size;
  }

  /* interpolate between the two extrems of the interval */
  /* p1 is the left and p2 is the right (so if p2 is before p1 we consider it's the next wave form (so the use of vx)) */
  x1=w->control_points_x[p1];
  y1=w->control_points_y[p1];
  x2=w->control_points_x[p2]+vx;
  y2=w->control_points_y[p2];

  /* linear interpolation - the easiest */
  /*
      x1,y1

                  sample_nb,y

		                     x2, y2

     (sample_nb-x1)/(x2-x1) = (y-y1)/(y2-y1)
     y= y1 + (y2-y1)*(sample_nb-x1)/(x2-x1)
   */

  y = y1 + (y2-y1)*((float)sample_nb-x1)/(x2-x1);

  return y;
}

int get_sample_sin(wave_form *w, int sample_nb)
{
  int y;
  int p1, p2, vx=0;
  register float x1, x2, y1, y2;
  register float per, ft, f;

  /* take the sample in the wave form */
  sample_nb %= w->x_size;

  /* search the good interval */
  p2=0;
  while (p2!=w->nb_control_points && w->control_points_x[p2]<=sample_nb) /* order of test is important */
    p2++;
  p1=p2-1;
  if (p2==w->nb_control_points) {
    p1=w->nb_control_points-1;
    p2=0;
    vx=w->x_size;
  }

  /* interpolate between the two extrems of the interval */
  /* p1 is the left and p2 is the right (so if p2 is before p1 we consider it's the next wave form (so the use of vx)) */
  x1=w->control_points_x[p1];
  y1=w->control_points_y[p1];
  x2=w->control_points_x[p2]+vx;
  y2=w->control_points_y[p2];

  /* linear interpolation - the easiest */
  /*
      x1,0

                  sample_nb,y

		                     x2, 1

     (sample_nb-x1)/(x2-x1) = y
     y= y1 + (y2-y1)*(sample_nb-x1)/(x2-x1)
   */
  per =((float)sample_nb-x1)/(x2-x1);
  ft = per*M_PI;
  f = (1.-cos(ft))*0.5;
  y = y1*(1.-f)+y2*f;

  return y;
}
#endif

int main(void)
{
  XEvent ev;
  SCREEN s;
  WAVE a, o;
  int i;

  if (!init_wave(&a, &o)) {
    fprintf(stderr, "Error with initing the sound\n");
    return 1;
  }

  /* init screen must come after init_wave ! */
  if (!init_screen(&s, &a, &o)) {
    fprintf(stderr, "Error with the interface - are you under X ?\n");
    return 1;
  }

  load_wave(s.a, "a.oa");
  draw_wave(&s, s.wf1, s.a);
  load_wave(s.o, "o.oa");
  draw_wave(&s, s.wf2, s.o);

  while(1) {
    register int sfd;

    /* barbarian but necessary, some event may be dequeued, so I might never see them if
     * I select when they are already out of the socket... deadlock ! */
    if (XCheckMaskEvent(s.d, -1L, &ev)!=False) {
      XPutBackEvent(s.d, &ev);
      if (handle_screen_event(&s)==-1)
	break;
    }

    sfd=get_sound_descriptor();
    i=wait_event(get_screen_descriptor(&s), sfd, is_mouse_in_play_window(&s) && sfd!=-1);

    if (i & SOUND_EVENT) {
      /* I should handle it better, but it's quick and fast like that (I should not access s fields, should I ?) */
      play_some_wave(&a, &o, s.my, s.mx, (int)s.mode4, (int)s.mode3, s.cf, s.no, s.dwb);
      s.mode3=(void *)s.mx;
      s.mode4=(void *)s.my;
    }

    if (i & SCREEN_EVENT)
      if (handle_screen_event(&s)==-1)
	break;
  }

  close_screen(&s);
  close_wave();
  return 0;
}
