/*
 * the oa project.
 *
 * Sed - december 1999
 *
 * This is public domain.
 *
 * the wave stuff file
 */

#include "waves.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "card.h"
#include "screen.h"
#include "wav.h"

/* return 0 if pb 1 if ok */
int init_wave(WAVE *a, WAVE *o)
{
  if (!(a->control_points_x=(int *)malloc(2*sizeof(int))) || !(a->control_points_y=(int *)malloc(2*sizeof(int)))) {
    perror("waves.init_wave malloc");
    return 0;
  }
  a->control_points_x[0]=0;
  a->control_points_y[0]=0;
  a->control_points_x[1]=256;
  a->control_points_y[1]=0;
  a->nb_control_points=2;
  a->x_size=512;

  if (!(o->control_points_x=(int *)malloc(2*sizeof(int))) || !(o->control_points_y=(int *)malloc(2*sizeof(int)))) {
    perror("waves.init_wave malloc");
    return 0;
  }
  o->control_points_x[0]=0;
  o->control_points_y[0]=0;
  o->control_points_x[1]=256;
  o->control_points_y[1]=0;
  o->nb_control_points=2;
  o->x_size=512;

  return 1;
}

#define NB_SAMPLE	2048

void play_some_wave(WAVE *a, WAVE *b, int y, int x, int oldy, int oldx, double cf, double no, int dump)
{
  static long double pos=0.0;
  register long double sa, so, per, perinc, inc, incinc;
  register long double freq;
  int i;
  short buf[NB_SAMPLE];

  per = (float)oldy/255.;
  perinc = (float)y/255.;
  perinc = (perinc-per)/NB_SAMPLE;

  freq=cf*pow(2., ((long double)oldx-512.)/(1024./no));
  inc = freq/(44100./512.);
  freq=cf*pow(2., ((long double)x-512.)/(1024./no));
  incinc = freq/(44100./512.);
  incinc=(incinc-inc)/NB_SAMPLE;

  if (pos>=1000000*512)
    pos-=1000000*512;

  while (not_full_sound()) {
    for (i=0; i<NB_SAMPLE; i++) {
      sa=get_sample_sin(a, pos);
      so=get_sample_sin(b, pos);

      buf[i]=so*per+sa*(1.-per);

      pos+=inc;
      inc+=incinc;
      per+=perinc;
    }
    play_sound(buf, NB_SAMPLE*2);
    if (dump==DOWN)
      dump_wav((char *)buf, NB_SAMPLE*2);
  }
}

int get_sound_descriptor(void)
{
  return get_sound_fd();
}

void close_wave(void)
{
}

long double get_sample_sin(WAVE *w, long double sample_nb)
{
  int y;
  int p1, p2, vx=0;
  register long double x1, x2, y1, y2;
  register long double per, ft, f;

  /* take the sample in the wave form */
  sample_nb=fmod(sample_nb, (long double)(w->x_size));

  /* search the good interval */
  p2=0;
  while (p2!=w->nb_control_points && w->control_points_x[p2]<=sample_nb) /* order of test is important */
    p2++;
  p1=p2-1;
  if (p2==w->nb_control_points) {
    p1=w->nb_control_points-1;
    p2=0;
    vx=w->x_size;
  }

  /* interpolate between the two extrems of the interval */
  /* p1 is the left and p2 is the right (so if p2 is before p1 we consider it's the next wave form (so the use of vx)) */
  x1=w->control_points_x[p1];
  y1=w->control_points_y[p1];
  x2=w->control_points_x[p2]+vx;
  y2=w->control_points_y[p2];

  /* linear interpolation - the easiest */
  /*
      x1,0

                  sample_nb,y

		                     x2, 1

     (sample_nb-x1)/(x2-x1) = y
     y= y1 + (y2-y1)*(sample_nb-x1)/(x2-x1)
   */
  per =(sample_nb-x1)/(x2-x1);
  ft = per*M_PI;
  f = (1.-cos(ft))*0.5;
  y = y1*(1.-f)+y2*f;

  return y;
}

/* return 0 if pb 1 if not */
int load_wave(WAVE *w, char *n)
{
  WAVE nw;
  char a[16], b[16], c[16];
  FILE *f;
  int i;

  if (!n || !strlen(n))
    return 0;

  f=fopen(n, "r");
  if (!f)
    return 0;

  if (fscanf(f, "%10s %10s %10s\n", a, b, c)!=3)
    goto l_error;

  if (strcasecmp(a, "oa") || strcasecmp(b, "data") || strcasecmp(c, "file"))
    goto l_error;

  if (fscanf(f, "%d", &nw.nb_control_points)!=1 || nw.nb_control_points<2)
    goto l_error;

  nw.control_points_x=(int *)malloc(nw.nb_control_points*sizeof(int));
  nw.control_points_y=(int *)malloc(nw.nb_control_points*sizeof(int));
  if (!nw.control_points_x || !nw.control_points_y)
    goto l_free_error;

  if (fscanf(f, "%d %d", &nw.control_points_x[0], &nw.control_points_y[0])!=2 || nw.control_points_x[0] ||
      nw.control_points_y[0]<-32768 || nw.control_points_y[0]>32767)
    goto l_free_error;

  for (i=1; i<nw.nb_control_points; i++) {
    if (fscanf(f, "%d %d", &nw.control_points_x[i], &nw.control_points_y[i])!=2)
      goto l_free_error;

    if (nw.control_points_x[i]<=nw.control_points_x[i-1] ||
	nw.control_points_x[i]>=512)
      goto l_free_error;

    if (nw.control_points_y[i]<-32768 || nw.control_points_y[i]>32767)
      goto l_free_error;
  }

  if (w->control_points_x)
    free(w->control_points_x);
  if (w->control_points_y)
    free(w->control_points_y);

  w->control_points_x=nw.control_points_x;
  w->control_points_y=nw.control_points_y;
  w->nb_control_points=nw.nb_control_points;
  w->x_size=512;

  fclose(f);
  return 1;

l_free_error:
  if (nw.control_points_x)
    free(nw.control_points_x);
  if (nw.control_points_y)
    free(nw.control_points_y);
l_error:
  fclose(f);
  return 0;
}

/* should check errors in there */
/* return 0 if pb 1 otherwise */
int save_wave(WAVE *w, char *n)
{
  FILE *f;
  int i;

  if (!n || !strlen(n))
    return 0;

  f=fopen(n, "w");
  if (!f)
    return 0;

  fprintf(f, "oa data file\n");
  fprintf(f, "%d\n", w->nb_control_points);
  for (i=0; i<w->nb_control_points; i++)
    fprintf(f, "%d %d\n", w->control_points_x[i], w->control_points_y[i]);

  fclose(f);

  return 1;
}
